"""
CARDAMOM - Benchmark paper version
========

Gene regulatory network inference from single-cell data
-------------------------------------------------------

Author: Elias Ventre
"""
__version__ = '1.0 - benchmark paper'

from .model import NetworkModel

__all__ = ['NetworkModel']